#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: Docker 配置生成器
@Author: Kermit
@Date: 2022-11-11 13:21:18
@LastEditors: Kermit
@LastEditTime: 2022-12-05 20:44:59
'''

import traceback
import os
from .config_loader import ConfigLoader
from .config import Algoinfo


def generate_docker_config(config_path: str):
    try:
        algorithm_config = ConfigLoader(config_path)
        algorithm_info = Algoinfo(algorithm_config.name, algorithm_config.version)

        gen_requirements_txt(algorithm_config.requirements)
        print('[AlgoSpace] [Step 1/4] Requirements has been successfully generated!')
        gen_dockerfile(algorithm_config.pre_command, algorithm_config.base_image, config_path)
        print('[AlgoSpace] [Step 2/4] Dockerfile has been successfully generated!')
        gen_docker_compose(algorithm_info.image_name, algorithm_info.image_version, algorithm_info.lower_name)
        print('[AlgoSpace] [Step 3/4] Docker compose file has been successfully generated!')
        gen_control_script(algorithm_info.image_name, algorithm_info.image_version)
        print('[AlgoSpace] [Step 4/4] Docker control file has been successfully generated!')
        print(f'[AlgoSpace] Generate successfully! Name: {algorithm_config.name}, Version: {algorithm_config.version}')
    except Exception as e:
        traceback.print_exc()
        print('[AlgoSpace] Generate error:', str(e))


def gen_requirements_txt(requirements: list[str]):
    with open('algospace-requirements.txt', 'w') as f:
        for requirements_item in requirements:
            if 'torch==' in requirements_item.replace(' ', ''):
                # 如果是Pytorch，要从官网下载可支持版本，否则会出问题，查看 Dockerfile 那里
                f.write('--find-links https://download.pytorch.org/whl/torch_stable.html\n')
                f.write(requirements_item + '\n')
            elif 'torchvision' in requirements_item:
                f.write('--find-links https://download.pytorch.org/whl/torch_stable.html\n')
                f.write(requirements_item + '\n')
            else:
                f.write(requirements_item + '\n')


def gen_dockerfile(pre_command: list[str], base_image: str, config_path: str):
    with open(os.path.join(os.path.split(__file__)[0], 'templates', 'docker', 'algospace-dockerfile'), 'r') as f:
        template = f.read()

    pre_command_lines = ''
    if len(pre_command) > 0:
        pre_command_lines += '''
RUN mv /etc/apt/sources.list /etc/apt/sources.list.bak && \\
    echo "deb http://mirrors.163.com/debian/ stretch main non-free contrib" >> /etc/apt/sources.list && \\
    echo "deb http://mirrors.163.com/debian/ stretch-updates main non-free contrib" >> /etc/apt/sources.list && \\
    echo "deb http://mirrors.163.com/debian/ stretch-backports main non-free contrib" >> /etc/apt/sources.list && \\
    echo "deb-src http://mirrors.163.com/debian/ stretch main non-free contrib" >> /etc/apt/sources.list && \\
    echo "deb-src http://mirrors.163.com/debian/ stretch-updates main non-free contrib" >> /etc/apt/sources.list && \\
    echo "deb-src http://mirrors.163.com/debian/ stretch-backports main non-free contrib" >> /etc/apt/sources.list && \\
    echo "deb http://mirrors.163.com/debian-security/ stretch/updates main non-free contrib" >> /etc/apt/sources.list && \\
    echo "deb-src http://mirrors.163.com/debian-security/ stretch/updates main non-free contrib" >> /etc/apt/sources.list && \\
    apt update && \\
'''
        for index, command_item in enumerate(pre_command):
            if index < len(pre_command) - 1:
                pre_command_lines += (f'    {command_item} && \\\n')
            else:
                pre_command_lines += (f'    {command_item}\n')
    template = template.replace('{PRE_COMMAND}', pre_command_lines)
    template = template.replace('{BASE_IMAGE}', base_image)
    template = template.replace('{CONFIG_PATH}', config_path)

    with open(os.path.join('algospace-dockerfile'), 'w') as f:
        f.write(template)


def gen_docker_compose(name: str, version: str, lower_name: str):
    with open(os.path.join(os.path.split(__file__)[0], 'templates', 'docker', 'algospace-docker-compose.yml'), 'r') as f:
        template = f.read()

    template = template.replace('{ALGORITHM_LOWER_NAME}', lower_name)
    template = template.replace('{ALGORITHM_NAME}', name)
    template = template.replace('{ALGORITHM_VERSION}', version)

    with open(os.path.join('algospace-docker-compose.yml'), 'w') as f:
        f.write(template)


def gen_control_script(name: str, version: str):
    with open(os.path.join(os.path.split(__file__)[0], 'templates', 'docker', 'algospace-docker-start.sh'), 'r') as f:
        template = f.read()
    with open(os.path.join('algospace-docker-start.sh'), 'w') as f:
        template = template.replace('{ALGORITHM_NAME}', name)
        template = template.replace('{ALGORITHM_VERSION}', version)
        f.write(template)

    with open(os.path.join(os.path.split(__file__)[0], 'templates', 'docker', 'algospace-docker-stop.sh'), 'r') as f:
        template = f.read()
    with open(os.path.join('algospace-docker-stop.sh'), 'w') as f:
        f.write(template)

    with open(os.path.join(os.path.split(__file__)[0], 'templates', 'docker', 'algospace-docker-logs.sh'), 'r') as f:
        template = f.read()
    with open(os.path.join('algospace-docker-logs.sh'), 'w') as f:
        f.write(template)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 
@Author: Ecohnoch(xcy)
@Date: 2022-10-06 12:30:47
@LastEditors: Kermit
@LastEditTime: 2022-11-11 19:45:41
'''

login_url = 'https://vsource.club/core/auth/login'



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 注册算法
@Author: Kermit
@Date: 2022-11-05 16:46:46
@LastEditors: Kermit
@LastEditTime: 2022-11-28 12:33:49
'''

from .config import enroll_url, verify_config_url, is_component_normal_url
import requests
import time
import traceback
from algospace.login import login, login_instance
from .config_loader import ConfigLoader


def enroll_from_config(config_path: str):
    ''' 从配置文件注册 '''
    try:
        algorithm_config = ConfigLoader(config_path)
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[AlgoSpace] Login...')
        if not login(algorithm_config.username, algorithm_config.password):
            print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                  '[AlgoSpace] Login failed. Please check your password.')
            return
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[AlgoSpace] Enroll processing...')
        enroll(algorithm_config.name, algorithm_config.version, algorithm_config.service_input, algorithm_config.service_output,
               algorithm_config.description, algorithm_config.scope, algorithm_config.chinese_name, algorithm_config.document, algorithm_config.config_file_content)
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
              f'[AlgoSpace] Enroll successfully! Name: {algorithm_config.name}, Version: {algorithm_config.version}')
    except Exception as e:
        traceback.print_exc()
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[AlgoSpace] Enroll error:', str(e))


def enroll(name: str, version: str, input: dict, output: dict, description: str, scope: str, chinese_name: str, document: str, config_file: str):
    ''' 注册算法 '''
    enroll_data = {
        'name': name,
        'version': version,
        'input_param': [{'key': key, 'type': val['type'], 'describe': val['describe']} for key, val in input.items()],
        'output_param': [{'key': key, 'type': val['type'], 'describe': val['describe']} for key, val in output.items()],
        'description': description,
        'scope': scope,
        'chinese_name': chinese_name,
        'document': document,
        'config_file': config_file
    }
    headers = login_instance.get_header()
    resp = requests.post(enroll_url, json=enroll_data, headers=headers)
    if resp.status_code != 200 and resp.status_code != 201:
        raise Exception(resp.status_code, resp.content.decode())
    if resp.json()['status'] != 200:
        raise Exception(resp.json().get('err_msg', 'Enroll algorithm error.'))
    return True


def verify_config(name: str, version: str, input: dict, output: dict):
    ''' 校验算法配置 '''
    enroll_data = {
        'name': name,
        'version': version,
        'input_param': [{'key': key, 'type': val['type'], 'describe': val['describe']} for key, val in input.items()],
        'output_param': [{'key': key, 'type': val['type'], 'describe': val['describe']} for key, val in output.items()]
    }
    headers = login_instance.get_header()
    resp = requests.post(verify_config_url, json=enroll_data, headers=headers)
    if resp.status_code != 200 and resp.status_code != 201:
        raise Exception(resp.status_code, resp.content.decode())
    if resp.json()['status'] != 200:
        raise Exception(resp.json().get('err_msg', 'Verify algorithm config error.'))
    file = resp.json()['data'].get('file')
    if file is not None:
        return file
    else:
        return ''


def is_component_normal(name: str, version: str):
    ''' 获取算法转发组件状态正常 '''
    headers = login_instance.get_header()
    resp = requests.get(f'{is_component_normal_url}?name={name}&version={version}', headers=headers)
    if resp.status_code != 200 and resp.status_code != 201:
        raise Exception(resp.status_code, resp.content.decode())
    if resp.json()['status'] != 200:
        raise Exception(resp.json().get('err_msg', 'Get component status error.'))
    is_component_normal = resp.json()['data'].get('is_component_normal')
    return is_component_normal



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 配置生成器
@Author: Kermit
@Date: 2022-11-07 17:30:04
@LastEditors: Kermit
@LastEditTime: 2022-12-03 21:09:03
'''

import os
import traceback


def generate_config():
    try:
        if os.path.exists('algospace-config.py'):
            print('[AlgoSpace] Init Error: \'algospace-config.py\' exists.')
            return
        with open(os.path.join(os.path.split(__file__)[0], 'templates', 'algospace-config.py'), 'r') as f:
            template = f.read()
        with open('algospace-config.py', 'w') as f:
            f.write(template)
        print('[AlgoSpace] Init successfully!')
        print('-----------------------------')
        print('已在当前目录下生成 \'algospace-config.py\' 文件，填写算法信息后，即可通过 \'algospace start\' 命令一键启动。')
    except Exception as e:
        traceback.print_exc()
        print('[AlgoSpace] Init error:', str(e))



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: AlgoSpace 脚本入口
@Author: Kermit
@Date: 2022-11-05 16:46:46
@LastEditors: Kermit
@LastEditTime: 2022-12-03 21:37:28
'''

from argparse import ArgumentParser


class ArgNamespace:
    command = 'command'
    version = 'version'
    config_path = 'config_path'
    enroll_remote = 'enroll_remote'
    generate_type = 'generate_type'


def run():
    parser = ArgumentParser('algospace', description='AlgoSpace 远程算法调用平台')
    subparsers = parser.add_subparsers(dest=ArgNamespace.command)  # ArgNamespace.command

    init_parser = subparsers.add_parser('init', description='初始化算法服务配置', help='初始化算法服务配置')
    start_parser = subparsers.add_parser('start', description='启动算法服务', help='启动算法服务')
    generate_parser = subparsers.add_parser('generate', description='生成容器配置', help='生成容器配置')
    enroll_parser = subparsers.add_parser('enroll', description='注册算法服务', help='注册算法服务')

    # 主命令参数
    parser.add_argument('-v', '--version', dest=ArgNamespace.version,
                        action='store_true', help='显示版本信息')  # ArgNamespace.version
    # start 命令参数
    start_parser.add_argument('-c', '--config', dest=ArgNamespace.config_path, metavar='<path>', type=str, required=False,
                              default='algospace-config.py', help='algospace-config.py 配置文件路径')  # ArgNamespace.config_path
    # generate 命令参数
    generate_parser.add_argument('-c', '--config', dest=ArgNamespace.config_path, metavar='<path>', type=str, required=False,
                                 default='algospace-config.py', help='algospace-config.py 配置文件路径')  # ArgNamespace.config_path
    # enroll 命令参数
    enroll_parser.add_argument('-c', '--config', dest=ArgNamespace.config_path, metavar='<path>', type=str, required=False,
                               default='algospace-config.py', help='algospace-config.py 配置文件路径')  # ArgNamespace.config_path
    # enroll_parser.add_argument('-r', '--remote', dest=ArgNamespace.enroll_remote, required=False,
    #                            action='store_true', default=False, help='远程部署')  # ArgNamespace.config_path

    args = ArgNamespace()
    parser.parse_args(namespace=args)

    if args.command is None or args.command == 'command':
        if args.version is True:
            from algospace import __version__
            print('AlgoSpace', __version__)
        else:
            parser.print_help()
    elif args.command == 'init':
        from algospace.provider.config_generator import generate_config
        generate_config()
    elif args.command == 'start':
        from algospace.provider.service import run_service
        run_service(args.config_path)
    elif args.command == 'generate':
        from algospace.provider.docker_generator import generate_docker_config
        generate_docker_config(args.config_path)
    elif args.command == 'enroll':
        from algospace.provider.enroll import enroll_from_config
        if not args.enroll_remote:
            enroll_from_config(args.config_path)
        else:
            print('Enrolling remotely will come soon.')
    else:
        raise Exception(f'CommandError: \'{args.command}\' is an invalid command')


if __name__ == '__main__':
    run()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 标准输入输出处理
@Author: Kermit
@Date: 2022-11-07 14:56:36
@LastEditors: Kermit
@LastEditTime: 2022-12-11 14:33:30
'''

import sys
import os
from typing import Callable
import time
from multiprocessing import Queue


class RedirectPrint:
    ''' 重定向标准输出 '''

    def __init__(self, write: Callable):
        self.write = write
        self.original_stdout_write = None

    def open(self):
        sys.stdout.close()
        sys.stdout.write = self.original_stdout_write  # type: ignore

    def close(self):
        self.original_stdout_write = sys.stdout.write
        sys.stdout.write = self.write  # type: ignore

    def __enter__(self):
        self.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.open()


class HiddenPrint:
    ''' 隐藏标准输出 '''

    def __init__(self):
        self.original_stdout = None

    def open(self):
        sys.stdout.close()
        sys.stdout = self.original_stdout

    def close(self):
        self.original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __enter__(self):
        self.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.open()


class GradioPrint(RedirectPrint):
    ''' 用以 Gradio Service 的标准输出 '''

    def __init__(self):
        def write(arg):
            if arg != '\n' and arg != '\r':
                self.original_stdout_write(
                    f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}] [Gradio Service] {str(arg)}')  # type: ignore
            else:
                self.original_stdout_write(arg)  # type: ignore

        super().__init__(write)


class QueueStdIO:
    ''' 输出到队列的 IO，用于子进程通过队列将 IO 操作重定向到父进程 '''

    def __init__(self, io_name: str, queue: Queue):
        if io_name == 'stdout':
            self.io_name = io_name
            self.io = sys.stdout
            sys.stdout = self
        elif io_name == 'stderr':
            self.io_name = io_name
            self.io = sys.stderr
            sys.stderr = self
        self.queue = queue

    def __getattr__(self, name: str):
        attr = getattr(self.io, name)
        if hasattr(attr, '__call__'):
            def call(*args, **kwargs):
                self.queue.put((self.io_name, name, args, kwargs))
            return call
        else:
            return self.io


class QueueStdIOExec:
    ''' 接收子进程 IO 操作并在父进程执行的处理器 '''

    def __init__(self, queue: Queue):
        self.queue = queue

    def exec(self):
        try:
            io_name, name, args, kwargs = self.queue.get(block=False)
            if io_name == 'stdout':
                io = sys.stdout
            elif io_name == 'stderr':
                io = sys.stderr
            else:
                raise Exception('Invalid io_name')
            attr = getattr(io, name)
            if hasattr(attr, '__call__'):
                attr(*args, **kwargs)
            if name == 'write':
                io.flush()
        except:
            pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 配置加载器
@Author: Kermit
@Date: 2022-11-05 20:19:06
@LastEditors: Kermit
@LastEditTime: 2022-12-12 18:42:35
'''

import os
import sys
import re
from typing import Callable, Optional
from algospace.exceptions import ConfigError

valid_param_type = ['str', 'int', 'float', 'image_path', 'video_path', 'voice_path']


class ConfigLoader:
    def __init__(self, config_path: str, is_verify_service: bool = False, is_verify_self_gradio_launch: bool = False) -> None:
        self.config_path = config_path
        self.is_verify_service = is_verify_service
        self.is_verify_self_gradio_launch = is_verify_self_gradio_launch
        # 导入配置
        config_dirpath = os.path.split(config_path)[0]
        config_filename = os.path.split(config_path)[1]
        config_filename_noext = os.path.splitext(config_filename)[0]
        sys.path.insert(0, config_dirpath)
        config = __import__(config_filename_noext)
        sys.path.pop(0)
        with open(config_path, 'r') as f:
            self.config_file_content = f.read()
            self.config_dirpath = config_dirpath

        self.name: str = getattr(config, 'name', '')
        self.version: str = getattr(config, 'version', '')
        self.username: str = getattr(config, 'username', '')
        self.password: str = getattr(config, 'password', '')
        self.service_filepath: str = getattr(config, 'service_filepath', '')
        self.service_function: str = getattr(config, 'service_function', '')
        self.service_input: dict = getattr(config, 'service_input')
        self.service_output: dict = getattr(config, 'service_output')
        self.service_max_parallel: int = int(getattr(config, 'service_max_parallel', 1))
        self.service_timeout: float = float(getattr(config, 'service_timeout', 60))
        self.service_tmp_dir: str = getattr(config, 'service_tmp_dir', '/tmp')

        self.description: str = getattr(config, 'description', '')
        self.scope: str = getattr(config, 'scope', 'PRIVATE')
        self.chinese_name: str = getattr(config, 'chinese_name', '')
        self.document_filepath: str = getattr(config, 'document_filepath', '')
        self.document: str = ''

        self.requirements: list[str] = getattr(config, 'requirements', [])
        self.pre_command: list[str] = getattr(config, 'pre_command', [])
        self.base_image: str = getattr(config, 'base_image', 'python:3.9')

        self.gradio_launch_filepath: str = getattr(config, 'gradio_launch_filepath', '')
        self.gradio_launch_function: str = getattr(config, 'gradio_launch_function', '')
        self.gradio_launch_host: str = getattr(config, 'gradio_launch_host', '127.0.0.1')
        self.gradio_launch_port: int = getattr(config, 'gradio_launch_port', 7860)
        self.is_self_gradio_launch = bool(self.gradio_launch_filepath and self.gradio_launch_function)

        if self.is_self_gradio_launch:
            self.gradio_server_host = self.gradio_launch_host
            self.gradio_server_port = self.gradio_launch_port
        else:
            self.gradio_server_host = '127.0.0.1'
            self.gradio_server_port = 7860

        self._verify()

    def _verify(self) -> None:
        if type(self.name) != str:
            raise ConfigError('ConfigError: \'name\' is not str.')
        if len(self.name) == 0:
            raise ConfigError('ConfigError: \'name\' is empty.')
        if not bool(re.match('^[a-zA-Z0-9_\-@]+$', self.name)):
            raise ConfigError('ConfigError: \'name\' can only include \'a-z A-Z 0-9 _ - @\'.')

        if type(self.version) != str:
            raise ConfigError('ConfigError: \'version\' is not str.')
        if len(self.version) == 0:
            raise ConfigError('ConfigError: \'version\' is empty.')
        if not bool(re.match('^[a-zA-Z0-9\.]+$', self.version)):
            raise ConfigError('ConfigError: \'version\' can only include \'a-z A-Z 0-9 .\'.')

        if type(self.username) != str:
            raise ConfigError('ConfigError: \'username\' is not str.')
        if len(self.username) == 0:
            raise ConfigError('ConfigError: \'username\' is empty.')

        if type(self.password) != str:
            raise ConfigError('ConfigError: \'password\' is not str.')
        if len(self.password) == 0:
            raise ConfigError('ConfigError: \'password\' is empty.')

        if type(self.service_filepath) != str:
            raise ConfigError('ConfigError: \'service_filepath\' is not str.')
        if len(self.service_filepath) == 0:
            raise ConfigError('ConfigError: \'service_filepath\' is empty.')
        if not os.path.exists(self._get_service_file_info()['service_filepath']):
            raise ConfigError(f'ConfigError: file \'{self.service_filepath}\' does not exist.')

        if type(self.service_function) != str:
            raise ConfigError('ConfigError: \'service_function\' is not str.')
        if len(self.service_function) == 0:
            raise ConfigError('ConfigError: \'service_function\' is empty.')
        if self.is_verify_service:
            self.verify_service()

        if type(self.service_input) != dict:
            raise ConfigError('ConfigError: \'service_input\' is not dict.')
        if len(self.service_input) == 0:
            raise ConfigError('ConfigError: \'service_input\' is empty.')
        for key, info in self.service_input.items():
            if type(key) != str:
                raise ConfigError(f'ConfigError: key \'{str(key)}\' in \'service_input\' is not str.')
            if type(info) != dict:
                raise ConfigError(f'ConfigError: value of \'{str(key)}\' in \'service_input\' is not dict.')
            if info.get('type', '') not in valid_param_type:
                raise ConfigError(
                    f'ConfigError: type of \'{str(key)}\' in \'service_input\' is not in {str(valid_param_type)}.')
            info['describe'] = str(info.get('describe', ''))

        if type(self.service_output) != dict:
            raise ConfigError('ConfigError: \'service_output\' is not dict.')
        if len(self.service_output) == 0:
            raise ConfigError('ConfigError: \'service_output\' is empty.')
        for key, info in self.service_output.items():
            if type(key) != str:
                raise ConfigError(f'ConfigError: key \'{str(key)}\' in \'service_output\' is not str.')
            if type(info) != dict:
                raise ConfigError(f'ConfigError: value of \'{str(key)}\' in \'service_output\' is not dict.')
            if info.get('type', '') not in valid_param_type:
                raise ConfigError(
                    f'ConfigError: type of \'{str(key)}\' in \'service_output\' is not in {str(valid_param_type)}.')
            info['describe'] = str(info.get('describe', ''))

        if type(self.service_max_parallel) != int:
            raise ConfigError('ConfigError: \'service_max_parallel\' is not int.')
        if self.service_max_parallel <= 0:
            raise ConfigError('ConfigError: \'service_max_parallel\' is less than 0.')

        if type(self.service_timeout) != float:
            raise ConfigError('ConfigError: \'service_timeout\' is not float.')
        if self.service_timeout <= 0:
            raise ConfigError('ConfigError: \'service_timeout\' is less than 0.')

        if type(self.service_tmp_dir) != str:
            raise ConfigError('ConfigError: \'service_tmp_dir\' is not str.')

        if type(self.description) != str:
            raise ConfigError('ConfigError: \'description\' is not str.')
        if type(self.scope) != str:
            raise ConfigError('ConfigError: \'scope\' is not str.')
        if self.scope not in ['PRIVATE', 'GROUP', 'INSTITUTION', 'PUBLIC']:
            raise ConfigError('ConfigError: \'scope\' is not in \'PRIVATE\',\'GROUP\',\'INSTITUTION\',\'PUBLIC\'.')
        if type(self.chinese_name) != str:
            raise ConfigError('ConfigError: \'chinese_name\' is not str.')
        if type(self.document_filepath) != str:
            raise ConfigError('ConfigError: \'document_filepath\' is not str.')
        if self.document_filepath and not self._read_document_file():
            raise ConfigError(f'ConfigError: \'{self.document_filepath}\' does not exist.')

        if type(self.requirements) != list:
            raise ConfigError('ConfigError: \'requirements\' is not list.')
        for requirements_item in self.requirements:
            if type(requirements_item) != str:
                raise ConfigError('ConfigError: Item in \'requirements\' is not str.')

        if type(self.pre_command) != list:
            raise ConfigError('ConfigError: \'pre_command\' is not list.')
        for pre_command_item in self.pre_command:
            if type(pre_command_item) != str:
                raise ConfigError('ConfigError: Item in \'pre_command\' is not str.')
        if type(self.base_image) != str:
            raise ConfigError('ConfigError: \'base_image\' is not str.')

        if self.is_self_gradio_launch:
            if type(self.gradio_launch_filepath) != str:
                raise ConfigError('ConfigError: \'gradio_launch_filepath\' is not str.')
            if not os.path.exists(self._get_gradio_launch_file_info()['gradio_launch_filepath']):
                raise ConfigError(f'ConfigError: file \'{self.gradio_launch_filepath}\' does not exist.')
            if type(self.gradio_launch_function) != str:
                raise ConfigError('ConfigError: \'gradio_launch_function\' is not str.')
            if self.is_verify_self_gradio_launch:
                self.verify_gradio_launch()

            if type(self.gradio_launch_host) != str:
                raise ConfigError('ConfigError: \'gradio_launch_host\' is not str.')
            if len(self.gradio_launch_host) == 0:
                raise ConfigError('ConfigError: \'gradio_launch_host\' is empty.')
            if type(self.gradio_launch_port) != int:
                raise ConfigError('ConfigError: \'gradio_launch_port\' is not int.')

    @property
    def fn(self) -> Callable:
        return self._get_service_fn()  # type: ignore

    def verify_service(self) -> None:
        if self._import_service_file() is None:
            raise ConfigError(
                f'ConfigError: \'{self.service_function}\' does not exist.')
        if not hasattr(self._get_service_fn(), '__call__'):
            raise ConfigError(
                f'ConfigError: \'{self.service_function}\' is not callable.')

    def _get_service_fn(self) -> Optional[Callable]:
        return getattr(self._import_service_file(), self.service_function, None)

    def _import_service_file(self):
        service_path = self._get_service_file_info()['service_path']
        service_filename_noext = self._get_service_file_info()['service_filename_noext']
        if service_path not in sys.path:
            sys.path.insert(0, service_path)
        return __import__(service_filename_noext)

    def _get_service_file_info(self) -> dict:
        service_path = os.path.join(self.config_dirpath, os.path.split(self.service_filepath)[0])
        service_filename = os.path.split(self.service_filepath)[1]
        service_filename_noext = os.path.splitext(service_filename)[0]
        service_filepath = os.path.join(service_path, service_filename)
        return {
            'service_path': service_path,
            'service_filename': service_filename,
            'service_filename_noext': service_filename_noext,
            'service_filepath': service_filepath,
        }

    @property
    def service_tmp_path(self) -> str:
        return os.path.join(self.config_dirpath, self.service_tmp_dir)

    def _is_document_file_exist(self) -> bool:
        document_filepath = os.path.join(self.config_dirpath, self.document_filepath)
        return os.path.exists(document_filepath)

    def _read_document_file(self) -> bool:
        if not self._is_document_file_exist():
            return False
        document_filepath = os.path.join(self.config_dirpath, self.document_filepath)
        with open(document_filepath, 'r') as f:
            self.document = f.read()
        return True

    @property
    def gradio_launch_fn(self) -> Callable:
        return self._get_gradio_launch_fn()  # type: ignore

    def verify_gradio_launch(self) -> None:
        if self._import_gradio_launch_file() is None:
            raise ConfigError(
                f'ConfigError: \'{self.gradio_launch_function}\' does not exist.')
        if not hasattr(self._get_gradio_launch_fn(), '__call__'):
            raise ConfigError(
                f'ConfigError: \'{self.gradio_launch_function}\' is not callable.')

    def _get_gradio_launch_fn(self) -> Optional[Callable]:
        return getattr(self._import_gradio_launch_file(), self.gradio_launch_function, None)

    def _import_gradio_launch_file(self):
        gradio_launch_path = self._get_gradio_launch_file_info()['gradio_launch_path']
        gradio_launch_filename_noext = self._get_gradio_launch_file_info()['gradio_launch_filename_noext']
        if gradio_launch_path not in sys.path:
            sys.path.insert(0, gradio_launch_path)
        return __import__(gradio_launch_filename_noext)

    def _get_gradio_launch_file_info(self) -> dict:
        gradio_launch_path = os.path.join(self.config_dirpath, os.path.split(self.gradio_launch_filepath)[0])
        gradio_launch_filename = os.path.split(self.gradio_launch_filepath)[1]
        gradio_launch_filename_noext = os.path.splitext(gradio_launch_filename)[0]
        gradio_launch_filepath = os.path.join(gradio_launch_path, gradio_launch_filename)
        return {
            'gradio_launch_path': gradio_launch_path,
            'gradio_launch_filename': gradio_launch_filename,
            'gradio_launch_filename_noext': gradio_launch_filename_noext,
            'gradio_launch_filepath': gradio_launch_filepath,
        }



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 算法函数类
@Author: Kermit
@Date: 2022-11-11 15:47:20
@LastEditors: Kermit
@LastEditTime: 2022-12-05 17:42:32
'''

from typing import Optional, Callable
from algospace.login import login_instance
import os
import json
import time
import requests
import traceback
from . import config
from algospace.exceptions import CallException, TimeOutException
from algospace.provider.config_loader import valid_param_type


class AlgoFunction:
    ''' 算法函数 '''

    def __init__(self, name: str, version: Optional[str] = None, timeout: int = config.call_timeout) -> None:
        self.name = name
        self.version = version
        self.timeout = timeout
        self.info = None
        pass

    def write_file(self, local_path: str) -> str:
        ''' 将本地 path 转换为 storage 返回的 path '''
        filename = os.path.split(local_path)[-1]
        with open(local_path, 'rb') as f:
            files = {'file': (filename, f.read())}

        info = self.get_info()
        upload_url = f'{config.storage_file_url}/{info["name"]}/{info["version"]}'
        response = requests.post(upload_url, files=files, headers=login_instance.get_header())
        if response.status_code != 200 and response.status_code != 201 and response.json()['status'] != 200:
            raise Exception('Failed to write file.')
        if type(response.json()['data']) != dict:
            raise Exception('Write file success but response has no \'data\'.')
        path = response.json()['data'].get('filepath')
        if path is None:
            raise Exception('Write file success but response has no \'filepath\'.')

        return path

    def convert_url(self, path: str) -> str:
        ''' 将 path 转换为 url '''
        upload_url = f'{config.storage_file_url}/{path}'
        return upload_url

    def get_input_type_class(self, type: str, ) -> Callable:
        ''' 获取处理每种类型的 Callable 对象 '''
        if type not in valid_param_type:
            raise Exception(f'Param type \'{type}\' is not available. Please try to update your algospace version.')
        elif type == 'str':
            return str
        elif type == 'int':
            return int
        elif type == 'float':
            return float
        elif type == 'image_path':
            return self.write_file
        elif type == 'video_path':
            return self.write_file
        elif type == 'voice_path':
            return self.write_file
        else:
            return str

    def get_output_type_class(self, type: str, ) -> Callable:
        ''' 获取处理每种类型的 Callable 对象 '''
        if type not in valid_param_type:
            raise Exception(f'Param type \'{type}\' is not available. Please try to update your algospace version.')
        elif type == 'str':
            return str
        elif type == 'int':
            return int
        elif type == 'float':
            return float
        elif type == 'image_path':
            return self.convert_url
        elif type == 'video_path':
            return self.convert_url
        elif type == 'voice_path':
            return self.convert_url
        else:
            return str

    def __call__(self, *args: list, **kwds: dict) -> dict:
        ''' 调用函数 '''
        info = self.get_info()
        can_call = info['can_call']
        input_param = info['input_param']
        output_param = info['output_param']
        if not can_call:
            raise Exception('This algorithm is not callable now because the provider closed the service.')

        kwargs = {}
        for index, arg in enumerate(args):
            if len(input_param) == index:
                break
            kwargs[input_param[index]['key']] = arg
        kwargs.update(kwds)
        for param in input_param:
            if param['key'] not in kwargs.keys():
                self.show_example()
                raise Exception('Arguments are not correct. Please refer to the example above.')
            param_key = param['key']
            param_type = param['type']
            kwargs[param_key] = self.get_input_type_class(param_type)(kwargs[param_key])

        response = requests.post(info['submit_url'], data=kwargs,
                                 headers=login_instance.get_header(), timeout=self.timeout)
        if response.status_code != 200 and response.status_code != 201:
            raise Exception(response.status_code, response.content.decode())
        if response.json()['status'] != 200:
            raise CallException(response.json().get('err_msg', 'Call algorithm function error.'))
        req_id = response.json()['data']['id']

        start_time = time.time()
        while True:
            if time.time() - start_time >= self.timeout:
                raise TimeOutException('Timeout.')
            response = requests.get(info['get_result_url'], params={
                'id': req_id}, headers=login_instance.get_header(), timeout=self.timeout)
            if response.status_code != 200 and response.status_code != 201 and response.json()['status'] != 200:
                time.sleep(config.interval)
                continue
            result_status = response.json()['data']['status']
            result = response.json()['data']['result']
            if result_status == 'error':
                raise CallException(
                    f'There is something wrong in algorithm provider service: {result.get("err_msg", "(no error message).")}')

            for param in output_param:
                param_key = param['key']
                param_type = param['type']
                result[param_key] = self.get_output_type_class(param_type)(result[param_key])
            return result

    def get_info(self) -> dict:
        ''' 获取算法信息 '''
        if not self.info:
            query_dict = {'name': self.name}
            if self.version:
                query_dict['version'] = self.version
            response = requests.get(config.algo_info_url, params=query_dict, headers=login_instance.get_header())
            if response.status_code != 200 and response.status_code != 201 and response.json()['status'] != 200:
                raise Exception('Failed to get info.')
            algo_list = response.json()['data']['list']
            if len(algo_list) == 0:
                raise Exception(f'No algorithm named {self.name + (" " + self.version if self.version else "")}.')
            algo = algo_list[0]
            self.info = {
                'name': algo['name'],
                'version': algo['version'],
                'can_call': algo['can_call'],
                'input_param': algo['input_param'],
                'output_param': algo['output_param'],
                'create_date': algo['create_date'],
                'example': algo['example'],
                'output_example': algo['output_example'],
                'submit_url': algo['submit_url'],
                'get_result_url': algo['get_result_url']
            }
        return self.info

    def show_example(self):
        ''' 展示示例 '''
        info = self.get_info()
        print('调用方式:')
        print(info['example'])
        print('')
        print('返回示例:')
        print(info['output_example'])
        print('')



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 
@Author: Ecohnoch(xcy)
@Date: 2022-10-06 12:30:47
@LastEditors: Kermit
@LastEditTime: 2022-11-27 19:40:06
'''

import json
import time
from typing import Optional
import requests
import traceback
import algospace.config as config
from algospace.exceptions import LoginError


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class LoginInstance(metaclass=Singleton):
    def __init__(self, cookie: str = '', timestamp: int = 0):
        self.cookie = cookie
        self.timestamp = timestamp
        self.username = ''
        self.password = ''

    def ever_logined(self) -> bool:
        return self.username != '' and self.password != ''

    def is_outdated(self) -> bool:
        curr_timestamp = int(time.time())
        return curr_timestamp - self.timestamp > 10 * 24 * 60 * 60

    def set_cookie(self, cookie: str, timestamp: int) -> None:
        self.cookie = cookie
        self.timestamp = timestamp

    def get_cookie(self) -> str:
        if not self.ever_logined():
            raise LoginError('LoginError: No login before.')
        if self.is_outdated():
            self.login()
        return self.cookie

    def get_header(self) -> dict:
        return {'Cookie': self.get_cookie()}

    def login(self, username: Optional[str] = '', password: Optional[str] = '') -> bool:
        self.username = username or self.username
        self.password = password or self.password
        try:
            login_url = config.login_url
            login_data = {
                'user_id': self.username,
                'password': self.password
            }
            response = requests.post(login_url, data=login_data)
            if response.status_code != 200 and response.status_code != 201:
                raise Exception(response.status_code, response.content.decode())
            if response.json()['status'] != 200:
                raise Exception(response.json().get('err_msg', 'Login error.'))
            self.set_cookie(response.headers['Set-Cookie'], int(time.time()))
            return True
        except Exception as e:
            traceback.print_exc()
            return False


login_instance = LoginInstance()
login = login_instance.login



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 算法提供者核心服务
@Author: Kermit
@Date: 2022-11-05 16:46:46
@LastEditors: Kermit
@LastEditTime: 2022-12-12 22:42:10
'''

from typing import Callable, Optional
from . import config
from .config import Algoinfo
import urllib.request
import gradio as gr
import asyncio
import multiprocessing
from multiprocessing.synchronize import Lock
from multiprocessing.connection import Connection
import threading
import concurrent.futures
import socket
import traceback
import requests
from algospace.login import login, login_instance
from .config_loader import ConfigLoader, valid_param_type
from .enroll import enroll, verify_config, is_component_normal
from .stdio import GradioPrint, QueueStdIO, QueueStdIOExec
import shutil
import json
import time
import os
import re


class FnService:
    ''' 模型函数 Service '''

    def __init__(self, algorithm_config: ConfigLoader) -> None:
        self.algorithm_config = algorithm_config

    async def handle(self, *args, **kwargs):
        ''' 处理函数请求 '''
        future = asyncio.get_event_loop().run_in_executor(None, lambda: self.algorithm_config.fn(*args, **kwargs))
        out = await asyncio.wait_for(future, timeout=self.algorithm_config.service_timeout)
        return out

    def launch_thread(self, fn_index: int, fn_req_queue: multiprocessing.Queue, fn_res_queue: multiprocessing.Queue) -> None:
        ''' 启动函数服务线程 '''
        while True:
            try:
                args, kwargs = fn_req_queue.get()
                print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                      f'[Fn Service] [{fn_index}] Begin to calculate')
                out = asyncio.run(self.handle(*args, **kwargs))
                print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                      f'[Fn Service] [{fn_index}] Complete.')
                fn_res_queue.put((out, None))
            except Exception as e:
                fn_res_queue.put((None, e))

    def launch(self,
               fn_req_queue_list: list[multiprocessing.Queue],
               fn_res_queue_list: list[multiprocessing.Queue]) -> None:
        ''' 启动函数服务 '''
        self.algorithm_config.verify_service()  # 校验函数配置并附带将函数 import 入进程
        with concurrent.futures.ThreadPoolExecutor() as pool:
            pool.map(self.launch_thread,
                     [i for i in range(len(fn_req_queue_list))],
                     [fn_req_queue_list[i] for i in range(len(fn_req_queue_list))],
                     [fn_res_queue_list[i] for i in range(len(fn_res_queue_list))])


def get_available_fn_queue(
        fn_lock_list: list[Lock],
        fn_req_queue_list: list[multiprocessing.Queue],
        fn_res_queue_list: list[multiprocessing.Queue]) -> tuple[Lock, multiprocessing.Queue, multiprocessing.Queue]:
    ''' 获取当前可用的函数锁、函数队列 '''
    for i in range(len(fn_lock_list)):
        if fn_lock_list[i].acquire(block=False):
            return fn_lock_list[i], fn_req_queue_list[i], fn_res_queue_list[i]
    raise Exception('No available fn queue.')


class ApiService:
    ''' 处理 Api 调用的 Service '''

    def __init__(self, algorithm_config: ConfigLoader, algorithm_info: Algoinfo) -> None:
        self.algorithm_config = algorithm_config
        self.algorithm_info = algorithm_info

    def read_file(self, path: str) -> str:
        ''' 将 storage 返回的 path 转换为本地 path '''
        file_url = self.algorithm_info.storage_file_url + '/' + path
        tmp_path = self.algorithm_config.service_tmp_path
        dir_path = os.path.dirname(os.path.join(tmp_path, path))
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        file_request = urllib.request.Request(url=file_url,  method='GET')
        headers = login_instance.get_header()
        for key, value in headers.items():
            file_request.add_header(key, value)
        file_bytes = urllib.request.urlopen(file_request).read()

        with open(os.path.join(tmp_path, path), 'wb') as f:
            f.write(file_bytes)
        local_path = os.path.join(tmp_path, path)

        return local_path

    def write_file(self, local_path: str) -> str:
        ''' 将本地 path 转换为 storage 返回的 path '''
        filename = os.path.split(local_path)[-1]
        with open(local_path, 'rb') as f:
            files = {'file': (filename, f.read())}

        upload_url = f'{self.algorithm_info.storage_file_url}/{self.algorithm_info.name}/{self.algorithm_info.version}'
        response = requests.post(upload_url, files=files, headers=login_instance.get_header())
        if response.status_code != 200 and response.status_code != 201 and response.json()['status'] != 200:
            raise Exception('Failed to write file.')
        if type(response.json()['data']) != dict:
            raise Exception('Write file success but response has no \'data\'.')
        path = response.json()['data'].get('filepath')
        if path is None:
            raise Exception('Write file success but response has no \'filepath\'.')

        return path

    def get_input_type_class(self, type: str, ) -> Callable:
        ''' 获取处理每种类型的 Callable 对象 '''
        if type not in valid_param_type:
            raise Exception(f'Param type \'{type}\' is not available.')
        elif type == 'str':
            return str
        elif type == 'int':
            return int
        elif type == 'float':
            return float
        elif type == 'image_path':
            return self.read_file
        elif type == 'video_path':
            return self.read_file
        elif type == 'voice_path':
            return self.read_file
        else:
            return str

    def get_output_type_class(self, type: str, ) -> Callable:
        ''' 获取处理每种类型的 Callable 对象 '''
        if type not in valid_param_type:
            raise Exception(f'Param type \'{type}\' is not available.')
        elif type == 'str':
            return str
        elif type == 'int':
            return int
        elif type == 'float':
            return float
        elif type == 'image_path':
            return self.write_file
        elif type == 'video_path':
            return self.write_file
        elif type == 'voice_path':
            return self.write_file
        else:
            return str

    def handle(self,
               input_info: dict,
               fn_lock_list: list[Lock],
               fn_req_queue_list: list[multiprocessing.Queue],
               fn_res_queue_list: list[multiprocessing.Queue]) -> dict:
        ''' 处理请求 '''
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[Api Service] Begin to handle.')
        params = {}
        for key, info in self.algorithm_config.service_input.items():
            params[key] = self.get_input_type_class(info['type'])(input_info[key])

        fn_lock = None
        try:
            fn_lock, fn_req_queue, fn_res_queue = get_available_fn_queue(
                fn_lock_list, fn_req_queue_list, fn_res_queue_list)
            fn_req_queue.put(((), params))
            out, e = fn_res_queue.get()
            if e is not None:
                raise e
        finally:
            if fn_lock is not None:
                fn_lock.release()

        output_info = {}
        for key, info in self.algorithm_config.service_output.items():
            output_info[key] = self.get_output_type_class(info['type'])(out[key])
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[Api Service] Complete.')
        return output_info

    def get_example(self):
        is_pure_name = re.match('^[a-zA-Z0-9_]*$', self.algorithm_config.name) is not None   # type: ignore
        if is_pure_name:
            example = f'algospace.{self.algorithm_config.name}'
        else:
            example = f'algospace.fn(\'{self.algorithm_config.name}\')'
        example += f'(' + ", ".join([f"{key}: {input_dict['type']}" for key,
                                     input_dict in self.algorithm_config.service_input.items()])+')'
        return example


class GradioService:
    ''' 处理 Gradio 的 Service '''

    def __init__(self, algorithm_config: ConfigLoader, algorithm_info: Algoinfo) -> None:
        self.algorithm_config = algorithm_config
        self.algorithm_info = algorithm_info

    def get_type_component(self, type: str, describe: str):
        ''' 获取处理每种类型的 Gradio component '''
        if type not in valid_param_type:
            raise Exception(f'Param type \'{type}\' is not available.')
        elif type == 'str':
            return gr.Textbox(placeholder=describe, label=describe)
        elif type == 'int':
            return gr.Number(precision=0, label=describe)
        elif type == 'float':
            return gr.Number(label=describe)
        elif type == 'image_path':
            return gr.Image(type='filepath', label=describe)
        elif type == 'video_path':
            return gr.Video(type='filepath', label=describe)
        elif type == 'voice_path':
            return gr.Audio(type='filepath', label=describe)
        else:
            return gr.Textbox(placeholder=describe)

    def launch(self,
               fn_lock_list: list[Lock],
               fn_req_queue_list: list[multiprocessing.Queue],
               fn_res_queue_list: list[multiprocessing.Queue],
               gradio_port_con: Connection) -> None:
        ''' 启动 Gradio 服务 '''
        def fn(*args):
            kwargs = {}
            for index, (key, _) in enumerate(self.algorithm_config.service_input.items()):
                kwargs[key] = args[index]

            fn_lock = None
            try:
                fn_lock, fn_req_queue, fn_res_queue = get_available_fn_queue(
                    fn_lock_list, fn_req_queue_list, fn_res_queue_list)
                fn_req_queue.put(((), kwargs))
                out, e = fn_res_queue.get()
                if e is not None:
                    raise e
            finally:
                if fn_lock is not None:
                    fn_lock.release()

            output_info = []
            for key, _ in self.algorithm_config.service_output.items():
                output_info.append(out[key])
            return output_info if len(output_info) > 1 else output_info[0]

        inputs = []
        for _, info in self.algorithm_config.service_input.items():
            inputs.append(self.get_type_component(info['type'], info['describe']))
        outputs = []
        for _, info in self.algorithm_config.service_output.items():
            outputs.append(self.get_type_component(info['type'], info['describe']))

        gr_interface = gr.Interface(
            title=self.algorithm_info.full_name,
            # description=self.algorithm_config.description,
            fn=fn,
            inputs=inputs,
            outputs=outputs,
        )

        self.check_port()
        threading.Thread(target=self.check_launched, args=(gradio_port_con,), daemon=True).start()

        with GradioPrint():
            gr_interface.launch(
                show_api=False,
                show_tips=False,
                favicon_path=None,
                inline=True,
                quiet=True,
                height=500,
                width=900,
                server_name=self.algorithm_config.gradio_server_host,
                server_port=self.algorithm_config.gradio_server_port
            )

    def check_port(self):
        ''' 检测可用端口 '''
        while (True):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            state = sock.connect_ex((self.algorithm_config.gradio_server_host,
                                    self.algorithm_config.gradio_server_port))
            sock.close()
            if state != 0:
                break
            self.algorithm_config.gradio_server_port += 1

    def check_launched(self, gradio_port_con: Connection):
        ''' 检测 Gradio 服务已经启动 '''
        while (True):
            time.sleep(1)
            try:
                html_text = requests.get(
                    f'http://{self.algorithm_config.gradio_server_host}:{self.algorithm_config.gradio_server_port}').text
                if len(html_text) == 0:
                    raise
                gradio_port_con.send(self.algorithm_config.gradio_server_port)
                gradio_port_con.close()
            except:
                pass

    def launch_self(self, gradio_port_con: Connection) -> None:
        ''' 启动自主编写的 Gradio 服务 '''
        self.algorithm_config.verify_gradio_launch()  # 校验自主实现的 Gradio 配置并附带将应用 import 入进程
        threading.Thread(target=self.check_launched, args=(gradio_port_con,), daemon=True).start()
        with GradioPrint():
            self.algorithm_config.gradio_launch_fn()  # 启动 Gradio 服务

    async def handle(self, input_info: dict) -> dict:
        ''' 处理请求 '''
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[Gradio Service] Begin to handle.')

        req_method = input_info['method']
        req_apipath = input_info['apipath']
        req_headers = input_info['headers']
        req_query = input_info['query']
        req_body = input_info['body']
        req_is_body_json = input_info['is_body_json']
        if req_is_body_json:
            req_body = json.loads(req_body)
        url = f'http://{self.algorithm_config.gradio_server_host}:{self.algorithm_config.gradio_server_port}/{req_apipath}{"?" + req_query if req_query != "" else ""}'

        res = None
        if req_method == 'GET':
            res = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.get(url))
        if req_method == 'POST' and req_headers['Content-Type'] == 'application/json':
            res = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.post(url, json=req_body, headers=req_headers))
        if req_method == 'POST' and req_headers['Content-Type'] != 'application/json':
            res = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.post(url, data=req_body, headers=req_headers))

        if res is not None:
            res_status_code = res.status_code
            res_headers = dict(res.headers)
            res_body = res.content.hex()
            result = {
                'status_code': res_status_code,
                'headers': res_headers,
                'body': res_body,
            }
        else:
            result = {}

        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[Gradio Service] Complete.')
        return result

    async def handle_init(self) -> dict:
        ''' 处理 Gradio 初始化请求 '''
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[Gradio Init] Begin to handle.')

        if not os.path.exists('./frontend'):
            gradio_path = '/'.join(gr.__file__.split('/')[:-1])
            gradio_frontend_path = os.path.join(gradio_path, 'templates/frontend')
            shutil.copytree(gradio_frontend_path, './frontend')
            os.remove('./frontend/index.html')  # 删除未渲染的 index.html

        async def upload_file(bashpath: str, subpath: str):
            ''' 递归上传文件夹里的文件 '''
            path = os.path.join(bashpath, subpath)
            if os.path.exists(os.path.join(path, '.DS_Store')):
                os.remove(os.path.join(path, '.DS_Store'))
            for filename in os.listdir(path):
                filepath = os.path.join(path, filename)
                if os.path.isdir(filepath):
                    await upload_file(bashpath, os.path.join(subpath, filename))
                else:
                    url = self.algorithm_info.gradio_upload_url + '/' + subpath + '/' + filename
                    file = open(filepath, 'r', encoding="UTF-8", errors='ignore')
                    res = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.post(url, files={'file': file}, headers=login_instance.get_header()))
                    file.close()

        await upload_file('./frontend', '.')  # 上传文件夹里的文件

        if not os.path.exists('./frontend/index.html'):
            html_res = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.get(
                f'http://{self.algorithm_config.gradio_server_host}:{self.algorithm_config.gradio_server_port}'))
            html_text = html_res.text
            html_file = open('./frontend/index.html', 'w+', encoding="UTF-8", errors='ignore')
            html_file.write(html_text)
            html_file.close()
        url = self.algorithm_info.gradio_upload_url + '/index.html'
        file = open('./frontend/index.html', 'r', encoding="UTF-8", errors='ignore')
        # 上传 index.html
        res = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.post(url, files={'file': file}, headers=login_instance.get_header()))
        file.close()

        shutil.rmtree('./frontend')  # 删除已上传的文件夹

        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[Gradio Init] Complete.')
        return {}


class Service:
    ''' 算法提供方 Service '''

    def __init__(
        self,
        config_path: str,
    ):
        self.algorithm_config = ConfigLoader(config_path)
        self.algorithm_info = Algoinfo(self.algorithm_config.name, self.algorithm_config.version)

        self.login_headers = {}
        self.login_headers_timestamp = 0

        self.fn_service = FnService(self.algorithm_config)
        self.api_service = ApiService(self.algorithm_config, self.algorithm_info)
        self.gradio_service = GradioService(self.algorithm_config, self.algorithm_info)

    def launch(self,
               stdio_queue: multiprocessing.Queue,
               type: str,
               fn_lock_list: list[Lock],
               fn_req_queue_list: list[multiprocessing.Queue],
               fn_res_queue_list: list[multiprocessing.Queue],
               gradio_port_con: Optional[Connection] = None):
        try:
            QueueStdIO('stdout', stdio_queue)
            QueueStdIO('stderr', stdio_queue)
            if not self.login():
                raise Exception('Login failed.')
            if type == 'FN':
                self.fn_service.launch(fn_req_queue_list, fn_res_queue_list)
            elif type == 'SERVICE':
                if not gradio_port_con:
                    raise Exception('Argument \'gradio_port_con\' is None.')
                self.launch_service(fn_lock_list, fn_req_queue_list, fn_res_queue_list, gradio_port_con)
            elif type == 'GRADIO':
                if not gradio_port_con:
                    raise Exception('Argument \'gradio_port_con\' is None.')
                if self.algorithm_config.is_self_gradio_launch:
                    self.gradio_service.launch_self(gradio_port_con)
                else:
                    self.gradio_service.launch(fn_lock_list, fn_req_queue_list, fn_res_queue_list, gradio_port_con)
            else:
                raise Exception('Argument "type" is unaccessable.')
        except Exception as e:
            traceback.print_exc()
            print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                  f'[{self.algorithm_info.upper_name}] Launch {type} error:', str(e))
            exit(1)

    def launch_service(self,
                       fn_lock_list: list[Lock],
                       fn_req_queue_list: list[multiprocessing.Queue],
                       fn_res_queue_list: list[multiprocessing.Queue],
                       gradio_port_con: Connection):
        # 从管道接受 Gradio 运行成功的端口
        gradio_port = int(gradio_port_con.recv())
        gradio_port_con.close()
        self.algorithm_config.gradio_server_port = gradio_port

        # 开始处理
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}] [Service] Your algorithm site: {self.algorithm_info.algorithm_site}')
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}] [Service] Start handling...')
        print('')

        async def start():
            parallel = {
                'max': self.algorithm_config.service_max_parallel,
                'curr': 0,
            }
            loop = asyncio.get_event_loop()
            while True:
                while parallel['curr'] >= parallel['max']:
                    # 并行数达到上限，等待
                    await asyncio.sleep(config.wait_interval)

                # 请求消息
                try:
                    req_info = await self.ask_data()
                    if req_info is None:
                        await asyncio.sleep(config.wait_interval)
                        continue
                except Exception as e:
                    traceback.print_exc()
                    print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                          '[Service] Ask data error:', str(e))
                    await asyncio.sleep(config.wait_interval)
                    continue

                # 处理请求
                try:
                    loop.create_task(self.handle(req_info, fn_lock_list,
                                     fn_req_queue_list, fn_res_queue_list, parallel))
                    await asyncio.sleep(config.call_interval)
                except Exception as e:
                    traceback.print_exc()
                    print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                          '[Service] Create handle task error:', str(e))
                    await asyncio.sleep(config.wait_interval)

        asyncio.run(start())

    async def ask_data(self):
        ask_for_data_resp = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.get(self.algorithm_info.ask_data_url, headers=login_instance.get_header()))
        if ask_for_data_resp.status_code != 200 and ask_for_data_resp.status_code != 201:
            raise Exception(ask_for_data_resp.status_code, ask_for_data_resp.content.decode())
        ask_for_data_dict = ask_for_data_resp.json()
        if ask_for_data_dict['status'] == 201:
            # 没有新的请求信息
            return
        if ask_for_data_dict['status'] != 200:
            raise Exception('the response', ask_for_data_dict)
        req_info = ask_for_data_dict.get('data', {})
        if type(req_info) != dict:
            raise Exception('the property of \'data\' is not dict.')
        req_info = req_info.get('req_info', {})
        if type(req_info) != dict:
            raise Exception('the property of \'req_info\' is not dict.')
        return req_info

    async def handle(self,
                     req_info: dict,
                     fn_lock_list: list[Lock],
                     fn_req_queue_list: list[multiprocessing.Queue],
                     fn_res_queue_list: list[multiprocessing.Queue],
                     parallel: dict):
        try:
            parallel['curr'] += 1
            try:
                # 根据请求类型处理
                if req_info['type'] == 'api':
                    # 处理计算请求
                    result = await asyncio.get_event_loop().run_in_executor(None, lambda: self.api_service.handle(req_info['data'], fn_lock_list, fn_req_queue_list, fn_res_queue_list))
                elif req_info['type'] == 'gradio':
                    # 处理 Gradio 请求
                    result = await self.gradio_service.handle(req_info['data'])
                elif req_info['type'] == 'gradio_init':
                    # 处理 Gradio 初始化请求
                    try:
                        parallel['curr'] += parallel['max']  # 初始化时要阻塞所有并行请求
                        result = await self.gradio_service.handle_init()
                    finally:
                        parallel['curr'] -= parallel['max']
                else:
                    result = {}

                # 回传数据
                res_info = {
                    'id': req_info['id'],
                    'type': req_info['type'],
                    'status': 'finished',
                    'owner': req_info['owner'],
                    'create_date': req_info['create_date'],
                    'result': result
                }
                return_ans_param = {'res_info': res_info}
                return_ans_resp = await asyncio.get_event_loop().run_in_executor(None, lambda: requests.post(self.algorithm_info.return_ans_url,
                                                                                                             json=return_ans_param, headers=login_instance.get_header()))

                if return_ans_resp.status_code != 200 and return_ans_resp.status_code != 201 and return_ans_resp.json()['status'] != 200:
                    err_msg = "Service Result Return Error."
                    raise Exception(err_msg)
                else:
                    print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                          '[Service] Return ans success.')
            except Exception as e:
                traceback.print_exc()
                # 回传失败数据
                res_info = {
                    'id': req_info['id'],
                    'type': req_info['type'],
                    'status': 'error',
                    'owner': req_info['owner'],
                    'create_date': req_info['create_date'],
                    'result': {'err_msg': str(e)}
                }
                return_error_param = {'res_info': res_info}
                await asyncio.get_event_loop().run_in_executor(None, lambda: requests.post(self.algorithm_info.return_err_url, json=return_error_param,
                                                                                           headers=login_instance.get_header()))
                print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                      '[Service] Return error ans success.')
        except Exception as e:
            traceback.print_exc()
            print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[Service] Handle error:', str(e))
        finally:
            parallel['curr'] -= 1

    def login(self):
        if not login(self.algorithm_config.username, self.algorithm_config.password):
            print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                  f'[{self.algorithm_info.upper_name}] Login failed. Please check your password.')
            return False
        return True

    def enroll(self):
        try:
            enroll(self.algorithm_config.name,
                   self.algorithm_config.version,
                   self.algorithm_config.service_input,
                   self.algorithm_config.service_output,
                   self.algorithm_config.description,
                   self.algorithm_config.scope,
                   self.algorithm_config.chinese_name,
                   self.algorithm_config.document,
                   self.algorithm_config.config_file_content)
            return True
        except Exception as e:
            print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                  f'[{self.algorithm_info.upper_name}] Enroll failed:', str(e))
            return False

    def verify_config(self):
        try:
            file = verify_config(self.algorithm_config.name,
                                 self.algorithm_config.version,
                                 self.algorithm_config.service_input,
                                 self.algorithm_config.service_output)
            if file:
                dirpath = os.path.dirname(self.algorithm_config.config_path)
                filepath = os.path.join(dirpath, 'algospace-config-origin.py')
                with open(filepath, 'w') as f:
                    f.write(file)
                print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                      f'[{self.algorithm_info.upper_name}] Content of \'{self.algorithm_config.config_path}\' is not same as the config enrolled before. The original config file is regenerated at \'{filepath}\'. Please use the original file to start. Or if you have modified the algorithm, use a new \'version\' to start.')
                return False
            return True
        except Exception as e:
            print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                  f'[{self.algorithm_info.upper_name}] Verify config failed:', str(e))
            return False

    def is_component_normal(self):
        while (True):
            if is_component_normal(self.algorithm_config.name, self.algorithm_config.version):
                break
            print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                  f'[{self.algorithm_info.upper_name}] Waiting for components started...')
            time.sleep(5)

    def send_heartbeat(self):
        body = {
            'algorithm_name': self.algorithm_info.name,
            'algorithm_version': self.algorithm_info.version
        }
        requests.post(self.algorithm_info.heartbeat_url, data=body, headers=login_instance.get_header())

    async def start(self):
        # 初始化状态
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
              f'[{self.algorithm_info.upper_name}] Initializing...')
        if not self.login():
            return
        if not self.enroll():
            return
        if not self.verify_config():
            return
        self.is_component_normal()

        # 开启子进程
        stdio_queue = multiprocessing.Queue(maxsize=0)  # 子进程通过队列将标准 IO 重定向到父进程
        fn_lock_list = [multiprocessing.Lock() for _ in range(
            self.algorithm_config.service_max_parallel)]  # 模型函数的锁列表
        fn_req_queue_list = [multiprocessing.Queue(maxsize=0) for _ in range(
            self.algorithm_config.service_max_parallel)]  # 模型函数的请求队列列表
        fn_res_queue_list = [multiprocessing.Queue(maxsize=0) for _ in range(
            self.algorithm_config.service_max_parallel)]  # 模型函数的响应队列列表
        gradio_port_con1, gradio_port_con2 = multiprocessing.Pipe()  # 传递 Gradio 端口的管道

        def create_fn_process():
            fn_process = multiprocessing.Process(target=self.launch,
                                                 args=(stdio_queue,
                                                       'FN',
                                                       fn_lock_list,
                                                       fn_req_queue_list,
                                                       fn_res_queue_list),
                                                 daemon=True)
            fn_process.start()
            return fn_process

        def create_service_process():
            service_process = multiprocessing.Process(target=self.launch,
                                                      args=(stdio_queue,
                                                            'SERVICE',
                                                            fn_lock_list,
                                                            fn_req_queue_list,
                                                            fn_res_queue_list,
                                                            gradio_port_con2),
                                                      daemon=True)
            service_process.start()
            return service_process

        def create_gradio_process():
            gradio_process = multiprocessing.Process(target=self.launch,
                                                     args=(stdio_queue,
                                                           'GRADIO',
                                                           fn_lock_list,
                                                           fn_req_queue_list,
                                                           fn_res_queue_list,
                                                           gradio_port_con1),
                                                     daemon=True)
            gradio_process.start()
            return gradio_process

        def check_process_alive(process: multiprocessing.Process):
            is_alive = process.is_alive()
            if not is_alive:
                process.kill()
            return is_alive

        fn_process = create_fn_process()
        service_process = create_service_process()
        gradio_process = create_gradio_process()

        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
              f'[{self.algorithm_info.upper_name}] Waiting for service launched...')

        # 开启事件循环
        async def alive_task():
            nonlocal fn_process
            nonlocal service_process
            nonlocal gradio_process
            while True:
                await asyncio.sleep(1)
                if not check_process_alive(fn_process):
                    raise Exception('Fn process is not alive. Please check error log above.')
                if not check_process_alive(service_process):
                    raise Exception('Service process is not alive')
                if not check_process_alive(gradio_process):
                    raise Exception('Gradio process is not alive')

        async def heartbeat_task():
            times = 0
            while True:
                try:
                    await asyncio.sleep(1)
                    if times % 10 == 0:
                        # 每 10 秒发送心跳
                        await asyncio.get_running_loop().run_in_executor(None, self.send_heartbeat)
                        times = 0
                    times += 1
                except Exception as e:
                    traceback.print_exc()
                    print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                          f'[{self.algorithm_info.upper_name}] Heartbeat error:', str(e))

        async def subprocess_stdio_task():
            stdio_exec = QueueStdIOExec(stdio_queue).exec
            while True:
                try:
                    await asyncio.get_running_loop().run_in_executor(None, stdio_exec)
                except Exception as e:
                    traceback.print_exc()
                    print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]',
                          f'[{self.algorithm_info.upper_name}] Handle subprocess stdio error:', str(e))

        # 当有任务抛出异常时，停止所有任务
        tasks = [alive_task(), heartbeat_task(), subprocess_stdio_task()]
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)
        for task in pending:
            task.cancel()
        for task in done:
            task.result()


def run_service(config_path: str) -> None:
    loop = asyncio.get_event_loop()
    try:
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[AlgoSpace] Init.')
        service = Service(config_path)
        loop.run_until_complete(service.start())
    except:
        traceback.print_exc()
    finally:
        loop.close()
        print(f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]', '[AlgoSpace] Exit.')



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 异常
@Author: Kermit
@Date: 2022-10-06 12:30:47
@LastEditors: Kermit
@LastEditTime: 2022-11-11 18:50:46
'''

class ConfigError(Exception):
    def __init__(self, err_msg):
        Exception.__init__(self, err_msg)

class FileExtNotValid(Exception):
    def __init__(self, err_msg):
        Exception.__init__(self, err_msg)


class TimeOutException(Exception):
    def __init__(self, err_msg):
        Exception.__init__(self, err_msg)


class CallException(Exception):
    def __init__(self, err_msg):
        Exception.__init__(self, err_msg)


class LoginError(Exception):
    def __init__(self, err_msg):
        Exception.__init__(self, err_msg)


class InvalidAlgorithmVersion(Exception):
    def __init__(self, err_msg):
        Exception.__init__(self, err_msg)


''' AlgoSpace 算法服务配置文件 '''

############################
# 必填选项:
#   - name:                 算法名，仅能包含英文字母、数字和 _ - @ 符号
#   - version:              算法版本，仅能包含英文字母、数字和 . 符号
#   - username:             登录名
#   - password:             密码
#   - service_filepath:     算法服务文件路径，相对于本文件路径
#   - service_function:     算法服务函数名
#   - service_input:        算法服务函数输入参数，字典类型
#       - .{arg}:           参数名
#       - .{arg}.type:      参数类型，包含: str (字符串), int (定点数), float (浮点数), image_path (图片本地路径), video_path (视频本地路径), voice_path (音频本地路径)
#       - .{arg}.describe:  参数简要描述
#   - service_output:       算法服务函数输出参数，字典类型，函数需以对应字典类型输出
#       - .{arg}:           参数名
#       - .{arg}.type:      参数类型，包含: str (字符串), int (定点数), float (浮点数), image_path (图片本地路径), video_path (视频本地路径), voice_path (音频本地路径)
#       - .{arg}.describe:  参数简要描述
#   - service_timeout:      算法服务超时时间，单位: 秒
#   - service_max_parallel: 算法服务最大同时处理请求数；若大于 1，则需要保证算法服务函数处理时的线程安全，否则请通过多实例部署算法
#   - service_tmp_dir:      算法服务临时文件存放文件夹的路径，相对于本文件路径，用于函数输入参数包含 image_path, video_path, voice_path 时存放临时文件
#
# 参数配置填写示例:
#   示例函数:
#       def example_fn(text, image):
#           # 这里是具体算法实现 #
#           return { 'result_text': ..., 'result_image': ... }
#   对应参数配置:
#       service_input = {
#           'text': {
#               'type': 'str',
#               'describe': '输入一段文字',
#           },
#           'image': {
#               'type': 'image_path',
#               'describe': '输入一张图片',
#           }
#       }
#       service_output = {
#           'result_text': {
#               'type': 'str',
#               'describe': '文字返回结果'
#           },
#           'result_image': {
#               'type': 'image_path',
#               'describe': '图片返回结果'
#           }
#       }
############################

name = 'example_name'
version = 'v1.0'
username = 'user'
password = '****'
service_filepath = './main.py'
service_function = 'example_fn'
service_input = {
    'arg0': {
        'type': 'str',
        'describe': '',
    }
}
service_output = {
    'arg0': {
        'type': 'str',
        'describe': ''
    }
}
service_timeout = 60
service_max_parallel = 1
service_tmp_dir = './tmp'

############################
# 算法详情选项，可在网页端进行修改:
#   - description:          描述算法的简介，简要介绍算法的要点，可选格式：纯文本、HTML
#   - scope:                算法可见范围，可选值：PRIVATE（自己可见）、GROUP（小组内可见）、INSTITUTION（机构内可见）、PUBLIC（公开）
#   - chinese_name:         中文名，将与算法名一同向用户展示
#   - document_filepath:    文档文件路径，详细介绍算法的要点，可选格式：纯文本、Markdown、HTML，相对于本文件路径，如：'./README.md'
############################

description = ''
scope = 'PRIVATE'
chinese_name = ''
document_filepath = ''

############################
# 容器式部署选项，用于 generate 命令生成 docker 相关配置:
#   - requirements: 依赖包，如 ['torch', 'numpy', 'matplotlib==3.6.2']
#   - pre_command:  容器用于配置环境的命令，构建镜像时依次执行，如 ['apt install -y libgl1-mesa-glx']
#   - base_image:   构建容器镜像的基础镜像，默认使用 python:3.9
############################

requirements = []
pre_command = []
base_image = 'python:3.9'

############################
# 需要自主实现更复杂的 Gradio 应用时，可使用以下选项替代自动生成的 Gradio 应用:
#   - gradio_launch_filepath: 自主实现的 Gradio 入口文件路径，相对于本文件路径，如：'./gradio_launch.py'
#   - gradio_launch_function: 自主实现的 Gradio 入口函数名，如：'gradio_launch'，入口函数应当不含任何参数
#   - gradio_launch_host:     自主实现的 Gradio 服务器地址，同 Gradio 的 GRADIO_SERVER_NAME 配置，默认：'127.0.0.1'
#   - gradio_launch_port:     自主实现的 Gradio 端口号，同 Gradio 的 GRADIO_SERVER_PORT 配置，默认：7860
#
# 提示：
#   同一版本的算法，不可有多种 Gradio 的实现代码；
#   若修改 Gradio 的实现代码，请以新版本号启动。
############################

gradio_launch_filepath = ''
gradio_launch_function = ''
gradio_launch_host = '127.0.0.1'
gradio_launch_port = 7860



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: Setuptools
@Author: Kermit
@Date: 2022-10-06 12:30:47
@LastEditors: Kermit
@LastEditTime: 2022-12-12 14:46:48
'''

from setuptools import setup, find_packages
from os import path

name = 'algospace'
with open(path.join(name, 'version.txt'), 'r') as f:
    version = f.read().strip()

setup(
    name=name,
    version=version,
    description=(
        'AlgoSpace: A platform for displaying algorithm achievements'
    ),
    long_description=open('README.md', encoding='utf8').read(),
    long_description_content_type='text/markdown',
    author='DBIIR',
    author_email='ckeming@outlook.com',
    maintainer='DBIIR',
    maintainer_email='ckeming@outlook.com',
    license='BSD License',
    packages=find_packages(where='.', include=(name,)),
    platforms=["all"],
    url='https://github.com/Algo-Space/algospace-pypi',
    project_urls={
        'Documentation': 'https://vsource.club/publish',
        'Homepage': 'https://vsource.club',
    },
    install_requires=[
        'setuptools>=3.0',
        'requests',
        'gradio>=3.0'
    ],
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries'
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'algospace = algospace:run',
            'asc = algospace:run'
        ]
    }
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 
@Author: Kermit
@Date: 2022-11-11 15:27:48
@LastEditors: Kermit
@LastEditTime: 2022-11-11 22:04:55
'''


from typing import Optional
from algospace.exceptions import LoginError
from algospace.login import login as ori_login
from .config import call_timeout
from .fn import AlgoFunction


def login(username: str, password: str) -> None:
    ''' 登录 '''
    if not ori_login(username, password):
        raise LoginError('Login failed. Please check your password.')


_fn_storage = {}


def fn(algorithm_name: str, algorithm_version: Optional[str] = None, timeout: int = call_timeout) -> AlgoFunction:
    ''' 获取算法函数实例 '''
    global _fn_storage
    fn = _fn_storage.get(algorithm_name, {}).get(algorithm_version)
    if not fn:
        fn = AlgoFunction(algorithm_name, algorithm_version, timeout)
        fn_version = {algorithm_version: fn}
        _fn_storage[algorithm_name] = fn_version
    else:
        fn.timeout = timeout
    return fn


def call(algorithm_name: str, algorithm_version: Optional[str] = None, timeout: int = call_timeout, kwargs: dict = {}, **kwds: dict) -> dict:
    ''' 直接调用算法函数 '''
    kwargs.update(kwds)
    return fn(algorithm_name, algorithm_version, timeout)(**kwargs)


def show_example(algorithm_name: str, algorithm_version: Optional[str] = None):
    fn(algorithm_name, algorithm_version).show_example()


def info(algorithm_name: str, algorithm_version: Optional[str] = None) -> dict:
    ''' 获取算法信息 '''
    info = AlgoFunction(algorithm_name, algorithm_version).get_info()
    return {
        'name': info['name'],
        'version': info['version'],
        'can_call': info['can_call'],
        'input_param': info['input_param'],
        'output_param': info['output_param'],
        'create_date': info['create_date'],
        'example': info['example'],
        'output_example': info['output_example'],
    }


def clear_storage():
    ''' 清除算法信息缓存 '''
    global _fn_storage
    _fn_storage = {}



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 
@Author: Kermit
@Date: 2022-10-24 14:45:40
@LastEditors: Kermit
@LastEditTime: 2022-12-12 22:49:30
'''

wait_interval = 1
call_interval = 0.1


enroll_url = 'https://vsource.club/core/algo/pypi/enroll'
verify_config_url = 'https://vsource.club/core/algo/pypi/verify'
is_component_normal_url = 'https://vsource.club/core/algo/pypi/is_component_normal'


class Algoinfo:
    def __init__(self, algorithm_name: str, algorithm_version: str):
        self.name = algorithm_name
        self.version = algorithm_version

    @property
    def full_name(self):
        return f'{self.name}_{self.version}'

    @property
    def lower_name(self):
        return self.full_name.replace('-', '_').replace('.',  '_').lower()

    @property
    def upper_name(self):
        return self.full_name.replace('-', '_').replace('.',  '_').upper()

    @property
    def image_name(self):
        return self.name.replace('@', '_').lower()  # 镜像的名称需要限制为 [a-z0-9], 其中可以出现的符号为 [-._]

    @property
    def image_version(self):
        return self.version.lower()  # 镜像的名称需要限制为 [a-z0-9], 其中可以出现的符号为 [-._]

    @property
    def heartbeat_url(self):
        return 'https://vsource.club/core/heartbeat/send'

    @property
    def ask_data_url(self):
        return f'https://vsource.club/algo/{self.lower_name}/service/ask_data'

    @property
    def return_ans_url(self):
        return f'https://vsource.club/algo/{self.lower_name}/service/return_ans'

    @property
    def return_err_url(self):
        return f'https://vsource.club/algo/{self.lower_name}/service/error_ans'

    @property
    def gradio_upload_url(self):
        return f'https://vsource.club/algo/{self.lower_name}/gradio_worker/upload'

    @property
    def storage_file_url(self):
        return 'https://vsource.club/core/storage/file'

    @property
    def gradio_page(self):
        return f'https://vsource.club/algo/{self.lower_name}/gradio/'

    @property
    def algorithm_site(self):
        return f'https://vsource.club/algorithm/{self.name}/{self.version}'



'''
AlgoSpace: A platform for displaying algorithm achievements
'''

import pkgutil

from algospace.__main__ import run
from algospace.customer.service import login
from algospace.customer.service import info
from algospace.customer.service import fn
from algospace.customer.service import show_example
from algospace.customer.service import call
from algospace.customer.service import clear_storage


def __getattr__(algorithm_name: str):
    return fn(algorithm_name)


__version__ = pkgutil.get_data(__name__, 'version.txt').decode('ascii').strip()  # type: ignore



#!/usr/bin/env python
# -*- coding:utf-8 -*-
'''
@Description: 
@Author: Kermit
@Date: 2022-10-24 14:45:40
@LastEditors: Kermit
@LastEditTime: 2022-11-28 13:09:57
'''


interval = 0.3
call_timeout = 30

algo_info_url = 'https://vsource.club/core/algo/pypi/info'
storage_file_url = 'https://vsource.club/core/storage/file'
